// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: ArgParser.c,v 1.74.206.5 2014/12/12 19:41:18 jcorbett Exp $
#include "ArgParser.h"
#include "EULA.h"
#include "Svg.h"
#include "../version.inc"

#include "Portability/export/Port_MsgDispatcher.h"
#include "Portability/export/Port_Env.h"
#include "Portability/export/Port_Version.h"
#include "Portability/export/Port_FileName.h"
#include "Portability/export/basutcmdarg.h"
#include "TUtilities/export/smart_ptr.hpp"
#include "TUtilities/export/regex.hpp"

#include <stdexcept>
#include <iomanip>
#include <set>
#include <fstream>
#include <sstream>

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
// Note: this gets used for the overall ivt version.
static const char rcsrev[] = "$Revision: 1.74.206.5 $";

class XCad_CadBlock;
class DeviceModel;

// Static members of class ArgParser

// The IVT report is written to this output file stream.
ofstream ArgParser::reportOut;

// Errors that are generated before the report header has been
// completed are buffered in this output string stream.
ostringstream ArgParser::errorBuf;

// Functions that write error messages write to this output
// stream. Initially this is bound to a string buffer. After the
// header has been written, this is reassigned to the report output
// file stream so errors are written as they are generated.
ostream* ArgParser::errorOut(&errorBuf);

// Report summary information is written to a summary buffer.
ostringstream ArgParser::summaryBuf;

// SVG diagram output is written to this output file stream.
ofstream ArgParser::drawingOut;

// Additional detail is written in verbose mode.
bool ArgParser::verbose;

// For routing-cost based isolation analysis, debugging is sometimes
// aided by exhaustive reporting of intermediate routing search
// results.
bool ArgParser::allPaths;

// Used for automatically numbering report sections.  Note: this is
// incremented before it is used, so that the current value indicates
// the number of the section most recently generated.
unsigned ArgParser::sectionNumber(0);

// Used for automatically numbering report subsections. Implementation
// is similar to ArgParser::sectionNumber above, except the subsection
// number is automatically reset when a new section is generated.
unsigned ArgParser::subsectionNumber(0);

// Pattern used by the command line argument parsing library.
const char* usage(
   APPNAMESTR
   " ("
   // Command line arguments for UCF checking
   "  [-device <device>] "                      // e.g. xc4vlx60
   "  [-package <package>] "                    // e.g. ff1152
   "  {-group <label> <area_group>}"            // R#accept_isolation_regions
   "  [-spacing <clbs:1,"
                 TOSTRING(SPACING_MAX) ">]"
   "  [-nopin]"
   "  [-pig <pin_isolation_groups[.pig]>]"
   "  [-output <output[" RPT "]>]"
   "  [-verbose]"
   "  <constraints[.ucf]> "
   ") | ("
   // Command line arguments for NCD checking
   "  {-group <label> <module>}"                // R#accept_isolation_regions
   "  [-faults <faults:2,"
                TOSTRING(FAULTS_MAX) ">]"
   "  [-output <output[" RPT "]>]"
   "  [-nopin]"
   "  [-verbose]"
   "  [-allpaths]"                              // Extra fault reporting
   "  <combined[.ncd]>"                         // R#accept_isolation_regions
   ") | ("
   // Command line arguments for path analysis
   "  [-path <fpga_path[.scr]>]"                // Path Diagnostic Analysis
   "  [-output <output[" RPT "]>]"
   "  {-group <label> <module>}"
   "  [-allpaths]"                              // Extra fault reporting
   "  <combined[.ncd]>"
   ") | ("
   // License Agreement
   "  [-license]"
   ") | ("
   // Version
   "  [-version]"
   ")"
   );

// Defined in device.inc which is included by DeviceModel.c
extern const char* devicePat;
extern const char* archNameList[];
extern const size_t archCount;

// List the supported architectures for error reporting.
const string ArgParser::SupportedArchitectures() const {
   string archNames;
   string sep("   ");
   for (size_t i(0); i < archCount; ++i) {
      archNames += sep + archNameList[i];
      sep = "\n   ";
   }
   return archNames;
}

// Pattern that should match all devices in archDeviceList.
const boost::regex supportedDevicePat(devicePat);

// Parse and process command line arguments.
ArgParser::ArgParser(int argc, const char * const * argv, Data& data)
   : data(data), cmdArgs(argc, argv, usage)
{
   startTime = UT_TIME::Clock();

   string rev(version);

   if (cmdArgs.gaveusage()) {
      Console("| ([-f <command_file>]) | ([-h])\n\n"
              "Xilinx Isolation Verification Tool "
              + rev + "\n");
      throw false;
   } else {                                              // R#report_provenance
      if (cmdArgs.findsw("-version")) {
         Console(rev);
         throw false;
      }

      // R#report_provenance - In all other cases, start by printing
      // the banner.
      string banner("Xilinx Isolation Verification Tool " + rev + "\n\n"
                    "Copyright (C) 2006-2014 Xilinx, Inc.  All rights "
                    "reserved.\n"
                    "  '" APPNAMESTR
                    " -h' displays a brief argument summary.\n"
                    "  '" APPNAMESTR
                    " -license' displays the license agreement.\n");
      Console(banner);
   }

   // The command parser requires all switches to be optional and the -license
   // switch case has no other required arguments, so here we must check if
   // no arguments were supplied.
   if (cmdArgs.iserror() || argc == 1) {
      throw false;
   }

   if (cmdArgs.findsw("-license")) {
      Console(string(eula));
      throw false;
   }

   if (argc == 2 && argv[1] == string("-f")) {
      FatalUsage(136, "The -f switch must be followed by a file name.");
      // In this context, fatal errors are buffered and there's no
      // report file. By doing a simple return to the caller, the
      // provenance that would have been added to the log is instead
      // echoed to the console along with the error message.
      return;
   }

   // With respect to command line argument processing, there are two
   // modes: UCF-mode and NCD-mode. In UCF-mode, the argument list
   // contains the name of a constraints file. In NCD-mode, the
   // argument list contains an NCD file. The term "combined" in the
   // argument list refers to the NCD file containing the complete
   // design. Before single NCD support, multiple NCD files were
   // required, one partial design for each isolated function, and one
   // of which was the combined design.

   // The primary variable points to a constant indicating the
   // argument name of the primary input file in the usage parameter
   // for the UT_CMDARGS object.
   const char* primary(
      NeedConstraintsCheck() ? "constraints" : "combined");

   // Get the output file name. Default to the
   // basename of the primary file.
   if (cmdArgs.findsw("-output")) {
      outFile = string(cmdArgs.getsw("-output")->getargtxt());
   } else {
      // If the output file name is not specified on the command line,
      // use the base name of the input (combined design or UCF file),
      // but write the report to the current working directory, not
      // necessarily the directory containing the input file.
      Port::FileName outFileName(cmdArgs.getarg(primary)->getargtxt());
      outFileName.SetExtension(RPT);
      outFile = outFileName.GetTail(); // the part after last slash
   }
   reportOut.open(outFile.c_str());
   if (!reportOut) {            // R#report_errors
      FatalUsage(17, "Unable to open output file " + outFile + ".\n"
                 "Check the file path syntax, the permissions settings \n"
                 "of the target directory, and the free space on "
                 "the device.");
      return;
   }

   if (NeedConstraintsCheck()) {
      // R#report_errors
      if (!cmdArgs.findsw("-device") || !cmdArgs.findsw("-package")) {
         FatalUsage(2, "Both device and package must be specified for\n"
                    "user constraints file (.ucf) checking.");
         return;
      }

      data.constraintsFile = cmdArgs.getarg(primary)->getargtxt();
      data.device = cmdArgs.getsw("-device")->getargtxt();
      // Quick sanity check on the device.
      if (!boost::regex_match(data.device, supportedDevicePat)) {
         FatalUsage(135, "Device " + data.device + " is not supported.\n"
                    "Supported architectures are:\n"
                    + SupportedArchitectures());
      }
      data.package = cmdArgs.getsw("-package")->getargtxt();
      for (Port::INT16 i = 0;; ++i) {
         const UT_CMDSWINFO* gsw(cmdArgs.getsw("-group", i));
         if (!gsw) break;
         string areaGroup(gsw->getarg("area_group")->getargtxt());
         string isolationGroup(gsw->getarg("label")->getargtxt());
         pair<Data::AreaGroupSet::const_iterator, bool>
            itDidInsert(data.unusedAreaGroupSet.insert(areaGroup));
         if (!itDidInsert.second) { // R#report_errors
            FatalUsage(65, "Area group " + areaGroup + " is associated "
                       "with more\n"
                       "than one isolation group. An area group cannot "
                       "appear\n"
                       "more than once in the command line arguments.");
            return;

         }
         data.constraintMap.insert(make_pair(areaGroup, isolationGroup));
      }
      if (cmdArgs.findsw("-spacing")) {
         data.spacingMin = cmdArgs.getsw("-spacing")->getargnum();
      } else {
         // spacingMin initialized in constructor for Data class.
      }

      // Get the pin isolation groups (PIG) file
      // name. Default to the basename of the primary file.
      if (cmdArgs.findsw("-pig")) {
         if (!NeedPinCheck()) { // R#report_errors
            FatalUsage(15,
                       "The -nopin option and the -pig option are\n"
                       "mutually exclusive. The -nopin option "
                       "suppresses pin\n"
                       "analysis and obviates the pin isolation group (PIG)\n"
                       "file.");
            return;
         }
         Port::FileName pigFileName(cmdArgs.getsw("-pig")->getargtxt());
         LoadPinIsolationGroupsTxt(pigFileName);
      } else {
         // It's probably OK to have no pin isolation groups, but we
         // should report it.
      }
   } else {
      // Get the combined NCD file containing the complete routing.
      // We'll need this to make CadFile instances for the isolated modules.
      const char* primaryFile(cmdArgs.getarg(primary)->getargtxt());
      try {
         boost::shared_ptr<CadFile> primaryCad(new CadFile(primaryFile));
         data.groupData.insert(make_pair(string(primary), primaryCad));
         DeviceModel* deviceModel(data.GetDeviceModel());
         primaryCad->SetDeviceModel(deviceModel);
         data.architecture = deviceModel->GetArchitectureName();
         data.device = deviceModel->GetPartName();

         // Quick sanity check on the device. See also
         // supportedDevicePat above.
         if (!boost::regex_match(data.device, supportedDevicePat)) {
            FatalUsage(77, "Architecture " + data.architecture
                       + " is not supported.\n"
                       "Supported architectures are:\n"
                       + SupportedArchitectures());
         }

         // Get the group labels and the group NCD files/module names.
         std::vector<XCad_CadBlock*> isolatedModules;
         for (Port::INT16 i = 0;; ++i) {
            const UT_CMDSWINFO* gsw(cmdArgs.getsw("-group", i));
            if (!gsw) break;
            string moduleOrNcd(gsw->getarg("module")->getargtxt());
            string group(gsw->getarg("label")->getargtxt());

            // If the module is a block path, not an NCD file, find
            // XCad_CadBlock corresponding to the module and add it to
            // a list.
            Port::FileName ncdPath(moduleOrNcd.c_str());
            if (ncdPath.GetExtension() != ".ncd") {
               // Assume moduleOrNcd is the user hierarchical name of an
               // isolated module.
               XCad_CadBlock*
                  block(primaryCad->GetBlockByUserHierName(moduleOrNcd));
               isolatedModules.push_back(block);
            }
         }
         // Now that we have the list of module blocks, we can create
         // the CadFile instances.
         for (Port::INT16 i = 0;; ++i) {
            const UT_CMDSWINFO* gsw(cmdArgs.getsw("-group", i));
            if (!gsw) break;
            string moduleOrNcd(gsw->getarg("module")->getargtxt());
            string group(gsw->getarg("label")->getargtxt());

            boost::shared_ptr<CadFile>
               cadFile(new CadFile(moduleOrNcd, primaryCad, isolatedModules));
            DeviceModel* groupDeviceModel(
               new DeviceModel(cadFile->GetView()->GetDevice()));
            const string groupArch(groupDeviceModel->GetArchitectureName());
            delete groupDeviceModel;
            if (groupArch != data.architecture) {
               throw runtime_error(
                  "Input Error:76: The device family specified in "
                  + string(primaryFile)
                  + "\ndoes not match the one specified in " + moduleOrNcd
                  + ".\nThe device in all design files must match.");
            }
            data.groupData.insert(make_pair(group, cadFile));
         }
      } catch (runtime_error& cadFileError) {
         Fatal(cadFileError.what());
         return;
      }

      if (NeedPathAnalysis()) {
         Port::FileName scriptPath(cmdArgs.getsw("-path")->getargtxt());
         LoadScript(scriptPath);
      }

      // R#report_errors - Check that we have at least two groups by
      // counting adjacent groups that don't match.
      if (GroupCount() < 2 && !NeedPathAnalysis()) {
         FatalUsage(1,
                    "At least two distinct -group command line "
                    "arguments are\n"
                    "required for isolation analysis.");
         return;
      }

      if (cmdArgs.findsw("-faults")) {
         data.faultsMax = cmdArgs.getsw("-faults")->getargnum();
      } else {
         // See constructor.
      }
   }

   // Put the diagram in the same directory as the output report.
   Port::FileName drawingPath(Port::FileName(outFile.c_str()));
   drawingPath.SetExtension("");
   data.drawingTitle = string(drawingPath.GetTail());
   drawingPath.SetExtension(svgFileExtension.c_str());
   string drawingFileName(drawingPath.GetTail());
   drawingOut.open(string(drawingPath).c_str());
   if (!drawingOut) {
      FatalUsage(105,
                 "Unable to open output file " + drawingFileName + ".\n"
                 "Check the file path syntax, the permissions settings \n"
                 "of the target directory, and the free space on "
                 "the device.");
      return;
   }

   verbose = (cmdArgs.findsw("-verbose") == TRUE) || NeedPathAnalysis();
   allPaths = cmdArgs.findsw("-allpaths") == TRUE;
}

// R#report_provenance - Write title, time, version, command line, etc. to
// the output report.
void ArgParser::ReportHeader() {
   try {
      // Title, date, and start time.
      {
         UT_TIME::TIME_T now = UT_TIME::Time(NULL);
         string date(UT_TIME::CTime(&now));
         if (date.find('\n') == date.size() - 1) {
            date.resize(date.size() - 1);
         }
         Report("Isolation Verification Report\n");
         Report("  Date: " + date);
      }

      // Version
      Report("  IVT Version: " + string(version));

      // Version of ISE libraries IVT was compiled against.
      Report("  ISE Build: " + string(iseBuild));

      // Same information, but internal nomenclature.
      Report("  ISE Development Version: " + string(iseRelease));

      // This is where the ISE dynamic libraries and device model
      // files come from at run-time.
      //
      // Note: if the XILINX environment variable is not set, getValue
      // will print
      //    The XILINX environment variable is not set or is empty.
      // in a static constructor and immediately call exit(). This
      // gives us no recourse to provide better error handling except
      // to write our own version of the offending code.
      string iseInstallDir(Port_Env::GetEnv("XILINX"));
      Report("  XILINX Directory: " + iseInstallDir);

      Report("  ISE Installed Version: "
             + string(Port::Version::getValue("version")));

      // Arguments
      {
         char* cl(cmdArgs.getcmdline());
         Report("  Command Line: " APPNAMESTR " " + string(cl));
         delete[] cl;
      }

      // Current directory
      string cwd(Port::FileName(Port::FileName(".").GetAbsolutePath())
                 .GetHead());
      Report("  Working Directory: " + cwd);

      // Output file
      Report("  Output File: " + outFile);

      if (NeedConstraintsCheck()) {
         // Part and device
         Report("  Part: " + data.device + "-" + data.package);
      } else {
         // Input design files

         // Look up the part and file name from the combined group.
         Data::GroupIt combined(data.groupData.find("combined"));
         if (combined != data.groupData.end()) {
            Report("  Part: " + data.GetDeviceModel()->GetPartName() + "\n");
            const string combinedFile(combined->second->GetFileName());
            Report("  Design: " + combinedFile);

            Section("Isolated Modules");

            for (Data::GroupIt groupIt(data.groupData.begin());
                 groupIt != data.groupData.end();
                 ++groupIt)
            {
               if (groupIt->first != "combined") {
                  const string moduleFile(groupIt->second->GetFileName());
                  if (moduleFile != combinedFile) {
                     Report("  Group " + groupIt->first + " file: "
                            + moduleFile);
                  } else {
                     Report("  Group " + groupIt->first + " module: "
                            + groupIt->second->GetModuleName());
                  }
               }
            }
            if (data.groupData.empty()) {
               Report("  No input designs were loaded.");
            }
         } else {
            ;
            // Didn't find the combined design. An error will be
            // generated shortly.
         }
      }
   } catch (...) {
      const string errorMessage(errorBuf.str());
      if (!errorMessage.empty()) { // R#report_errors
         // Croak after writing as much header information as possible.
         Report(errorMessage);
         throw false;
      }
   }
   const string errorMessage(errorBuf.str());
   if (!errorMessage.empty()) { // R#report_errors
      // Croak after writing as much header information as possible.
      Report(errorMessage);
      throw false;
   }

   // Now that the header has been output, allow errors to be
   // written directly.
   errorOut = &reportOut;
}

// Check a UCF file?
bool ArgParser::NeedConstraintsCheck() {
   return !!cmdArgs.getarg("constraints");
}

// R#check_io_banks - Check the pins, IO buffers, and IO banks?
bool ArgParser::NeedPinCheck() {
   return !cmdArgs.findsw("-nopin");
}

// Read a path and report cost function analysis of it?
bool ArgParser::NeedPathAnalysis() {
   return !!cmdArgs.findsw("-path");
}

// R#report_errors - Write the elapsed time to the output report.
void ArgParser::ReportFooter() {

   Section("Isolation Verification Summary");

   string summaryStr(summaryBuf.str(), 1); // Skip extra initial linefeed.
   Report(summaryStr);

   if (!data.unroutedCount && !data.omittedPS7) {
      Report("\n  Isolation analysis completed.\n");
   } else {                     // R#report_errors
      Report("\n  Provisional isolation analysis completed.\n");
      if (data.unroutedCount) {
         Report("Input Error:62: The combined input design was not fully placed"
                " and routed.\n"
                "Review the implementation report and correct any\n"
                "errors preventing the design from being fully implemented.\n");
      } 
      if (data.omittedPS7) {
         Report("Input Error:147: The Zynq Processing Subsystem site was not"
                " included\n"
            "in the area range constraints (floorplan). Xilinx does not\n"
            "support this configuration mode in production.");
      }
   }
   // R#report_errors - Report run time. This will help us catch
   // unexpected performance changes and help us access the efficiency
   // of alternate implementations.
   long elapsedSec = (UT_TIME::Clock() - startTime)/UT_TIME::ClocksPerSec;
   ostringstream out;
   int hours   = elapsedSec / (60 * 60);
   int minutes = elapsedSec / 60 % 60;
   int seconds = elapsedSec % 60;
   out << "  Elapsed time: "
       << hours << ":"
       << setfill('0') << setw(2) << minutes << ":"
       << setfill('0') << setw(2) << seconds;
   Report(out.str());
}

// Explicitly close files to add debugging crashes that occur in the
// destructors that run just prior to exiting. This method is static
// so that it can be called in catch clauses that do not have access
// to "this".
void ArgParser::CloseFiles() {
   reportOut.close();
   drawingOut.close();
}

// R#report_errors - Count the groups by counting runs of each group
// in the sorted enumeration of groupData.
//
// We expect at least one group labeled "combined" because the
// argument parser won't let us get this far otherwise. Additional
// groups labeled "combined" indicate that the user selected
// "combined" as the name of an isolation group and that is not
// allowed.
//
// It is not allowed to label an isolation group "ignored". We expect
// no groups labeled "ignored" until later when nets are
// categorized.
int ArgParser::GroupCount() {
   size_t groupCount(0);
   for (Data::GroupIt gIt(data.groupData.begin());
        gIt != data.groupData.end();
        gIt = data.groupData.upper_bound(gIt->first))
   {
      ++groupCount;
   }
   string reserved;
   size_t combinedCount(data.groupData.count("combined"));
   groupCount -= combinedCount;
   size_t ignoredCount(data.groupData.count("ignored"));
   if (1 < combinedCount) {
      reserved = "combined";
   } else if (0 < ignoredCount) {
      reserved = "ignored";
   }
   if (!reserved.empty()) {
      FatalUsage(18, "The word \"" + reserved
                 + "\" cannot be used as a group label.\n"
                 "The words 'combined' and 'ignored' are reserved.\n");
      groupCount = 2;           // One error at a time.
   }
   return groupCount;
}

// Path loader for Path Diagnostic Analysis mode. See
// NetChecker::AnalyzePath().

const string sp("\\s*");        // match optional whitespace

const string nodePat("\\S|\\S.*\\S");

// submatch wire name, e.g.,
//     select -k wire 'OMUX(11185,-64136)'
const boost::regex wirePat(
   sp + "select" + sp + "-k" + sp + "wire"
   + sp + "'(" + nodePat + ")'" + sp,
   boost::regex::normal | boost::regbase::icase);

// submatch wire name, e.g.,
//     select -k arc 'OMUX(11185,-64136)--->HUNIHEX(6033,-65168)'
const boost::regex arcPat(
   sp + "select" + sp + "-k" + sp + "arc" + sp
   + "'(" + nodePat + ")--->(" + nodePat + ")'" + sp,
   boost::regex::normal | boost::regbase::icase);

// Load an FPGA Editor script and extract all the wire names from wire
// selection commands.
void ArgParser::LoadScript(const char* scriptPath) {
   ifstream script(scriptPath);
   if (script) {
      bool wantFirst(true);     // Becomes false after the first arc pushed.
      string wireName;          // AKA node name
      string nodeA, nodeB;      // a pair of nodes from an arc
      do {
         string line;
         getline(script, line, '\n');
         boost::smatch result;
         if (boost::regex_match(line, result, arcPat)) {
            if (result.size() != 3) { // R#report_errors
               FatalInternal(36, "Found script line that matches arcPat"
                             " but result.size() is " + ToString(result.size())
                             + ", not 3.");
               return;
            }
            nodeA = result.str(1);
            nodeB = result.str(2);
         } else if (boost::regex_match(line, result, wirePat)) {
            if (result.size() != 2) { // R#report_errors
               FatalInternal(37, "Found script line that matches nodePat"
                             " but result.size() is " + ToString(result.size())
                             + ", not 2.");
               return;
            }
            wireName = result.str(1);
            if (wantFirst) {
               if (nodeA == wireName) {
                  data.scriptWires.push_back(nodeB);
                  data.scriptWires.push_back(nodeA);
               } else {
                  data.scriptWires.push_back(nodeA);
                  data.scriptWires.push_back(nodeB);
               }
               wantFirst = false; // Got first. No longer want first.
            } else {
               data.scriptWires.push_back(wireName);
            }
         }
      } while (script);
      if (nodeA == wireName) {
         data.scriptWires.push_back(nodeB);
      } else if (nodeB == wireName) {
         data.scriptWires.push_back(nodeA);
      } else {                  // R#report_errors
         FatalInternal(13, "Nodes (aka wires) from last arc "
                       "do not contain last node in path.");
      }
   } else {                     // R#report_errors
      FatalInput(16, "Unable to load " + string(scriptPath) + ".\n"
                 "Check the file path, and file permissions.");
   }
}

// Pin Isolation Group loader

const string qq("[\"`']");      // match a quote " ` '
const string any(".*");         // match any string
const string sep("\\s+");       // match at least one whitespace character
const string comment("#.*");    // match a single line comment
const string noSpace("\\S+");
const string eol("$");

// Match a net name. This is tricky because slashes are hierarchy
// separators and also part of C-style /* */ comments and asterisks
// may appear in net names as wild-card indicators.
//
// The idea is to match all characters except ones that are not
// allowed in a net name.  Comment characters and quotation marks of
// various sorts would end a net name. If a forward slash occurs, it
// must be followed by something other than a *, a quote, or a comment
// symbol.
//
// This string is used as part of a pattern that matches surrounding 
// quotes explicitly.
const string noCommentNoQuote("(?:[^#/'`\"]|/(?:[^*#/'`\"]))+");

// This string is used for net names that are not quoted. When quotes
// are omitted, spaces are not allowed.
const string 
noCommentNoQuoteNoSpace("(?:[^#/'`\"[:space:]]|/(?:[^*#/'`\"[:space:]]))+");

// match blank or comment
const boost::regex discardPat("" + sp + "|" + sp + "#.*",
                              boost::regex::normal | boost::regbase::icase);

// ISOLATION_GROUP group_name BEGIN
// Submatch 2 is group_name
const boost::regex groupPat(sp + "ISOLATION_GROUP"
                            + sep
                            + "(" + noSpace + ")"
                            + sep + "BEGIN"
                            + "($|" + sep + any + ")",
                            boost::regex::normal | boost::regbase::icase);

// Match these:
//   END
//   END#aaa
//   END #aaa
//   END ISOLATION_GROUP
//   END ISOLATION_GROUP aaa
const boost::regex endPat(sp + "END"
                          + "("
                          +    eol
                          +   "|" + sp + comment
                          +   "|" + sep + "ISOLATION_GROUP" + eol
                          +   "|" + sep + "ISOLATION_GROUP" + sep + any
                          + ")",
                          boost::regex::normal | boost::regbase::icase);

// Match a single-line C style comment. E.g. /* foo */
const string cComment("(?:" + sp + "/\\*" + any + "\\*/" + sp + ")");
const string sepOrCComment("(?:" + sep + "|" + cComment + ")");

// Match these:
//   NET "net_name"
//   INST "net_name"
//   NET "net_name"aaa
//   INST "net_name"aaa
// Submatch 2 is net_name
const boost::regex netPat1(sp + "(NET|INST)" + sp
                          + qq + "(" + noCommentNoQuote + ")" + qq + any,
                          boost::regex::normal | boost::regbase::icase);
// Match these:
//   NET net_name
//   INST net_name
//   NET/* asdf */net_name"aaa
//   INST net_name aaa
// Submatch 2 is net_name
const boost::regex netPat2(sp + "(NET|INST)" + sepOrCComment
                           + "(" + noCommentNoQuoteNoSpace + ")"
                           + sepOrCComment + any,
                           boost::regex::normal | boost::regbase::icase);


// Read a file with net names that correspond to pins.  The net names
// are bracketed by isolation group names.
// Note: this is called before the header is printed, so it adds to
// errorBuf rather than calling Fatal().
void ArgParser::LoadPinIsolationGroupsTxt(const char* pigPath) {
   long lineCounter(0);
   set<Data::IsolationGroupName> groupSet;
   ifstream pig(pigPath);
   if (pig) {
      Data::IsolationGroupName group("<unspecified>");
      do {
         string line;
         getline(pig, line, '\n');
         ++lineCounter;
         boost::smatch result;
         if (boost::regex_match(line, discardPat)) {
            // Skip blank lines and comment lines.
            ;
         } else if (boost::regex_match(line, result, groupPat)) {
            if (group != "<unspecified>") { // R#report_errors
               FatalInput(69, "Unclosed isolation group in pin isolation\n"
                          "group file at line " + ToString(lineCounter) + ".\n"
                          "Missing END statement.");
               return;
            }
            group = result.str(1);
            // Verify we've heard of this group
            bool foundIt(false);
            for (Data::ConstraintMapIt cmIt(data.constraintMap.begin());
                 cmIt != data.constraintMap.end();
                 ++cmIt)
            {
               if (cmIt->second == group) {
                  foundIt = true;
                  break;
               }
            }
            if (!foundIt) {     // R#report_errors
               // Arguments in the command line arguments don't match
               // the PIG file.
               if (data.constraintMap.find(group) == data.constraintMap.end())
               {
                  FatalInput(56, "Isolation group " + group
                             + " in pin isolation group file\n"
                             "was not defined in the arguments.");
               } else {
                  // Give an extra hint if the putative isolation group name
                  // happens to be an area group name.
                  FatalInput(56, "Isolation group " + group
                             + " in pin isolation group file\n"
                             "was not defined in the arguments."
                             "\nGroup " + group + " is an area group name.");
               }
            }
            bool isNew(groupSet.insert(group).second);
            if (!isNew) {       // R#report_errors
               FatalInput(57, "Group " + group + " was redefined in\n"
                          "pin isolation group file at line "
                          + ToString(lineCounter) + ".\n"
                          "Each isolation group must only be defined once.");
               return;
            }
         } else if (group != "<unspecified>" &&
                    (boost::regex_match(line, result, netPat1)
                     || boost::regex_match(line, result, netPat2)))
         {                      // R#report_errors
            if (result.size() != 3) {
               FatalInternal(35, "Line " + ToString(lineCounter)
                             + " in pin isolation group file\n"
                             "matches netPat but has the wrong number "
                             "of fields.");
               return;
            }
            string net(result.str(2));
            bool isNew(data.pinGroupMap.insert(make_pair(net, group)).second);
            if (!isNew) {
               FatalInput(58, "Duplicate pin net name\n"
                          + net + " found in\n"
                          "pin isolation group file at line "
                          + ToString(lineCounter) + ".\n"
                          "Each pin must be listed at most once.");
               return;
            }
         } else if (boost::regex_match(line, result, endPat)
                    && group != "<unspecified>")
         {
            group = "<unspecified>";
         } else {               // R#report_errors
            if (group == "<unspecified>") {
               FatalInput(67, "Expecting an isolation group in pin\n"
                          "isolation group file at line "
                          + ToString(lineCounter) + ".\n"
                          "Check for a missing semicolon, "
                          "quotation mark or comment delimiter.");
               return;
            } else {
               FatalInput(68, "Expecting a net name in pin\n"
                          "isolation group file at line "
                          + ToString(lineCounter) + ".\n"
                          "Check for a missing semicolon, "
                          "quotation mark or comment delimiter.");
               return;
            }
         }
      } while (pig);
      if (group != "<unspecified>") { // R#report_errors
         FatalInput(102, "Unclosed isolation group in pin\n"
                    "isolation group file at line "
                    + ToString(lineCounter) + ".\n"
                    "Check for a missing semicolon, "
                    "quotation mark or comment delimiter.");
      }
   } else {                     // R#report_errors
      FatalInput(55, "Unable to load " + string(pigPath) + ".\n"
                 "Check the file path, and file permissions.");
   }
}
