// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: NetChecker.c,v 1.51.206.2 2013/05/10 00:04:06 jcorbett Exp $

#include "JDBHelpers/Pds/export/Pdl/Pds_PdlPin.h"
#include "JDBHelpers/Pds/export/Pdl/Pds_PdlExtendedPinType.h"
#include "TUtilities/export/smart_ptr.hpp"

#include "DeviceModel.h"
#include "ArgParser.h"

#include "NetChecker.h"

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.51.206.2 $";

// R#check_irs_separation

NetChecker::NetChecker(Data& cadData)
   : data(cadData), faultList(cadData.netGroup),
     faultCount(0), pathCount(0), group1(0), group2(0)
{
   if (cadData.architecture == "virtex4") {
      costFunction.reset(new Virtex4Cost());
      ReportUnroutedNets();
      LabelNodes();
   } else {
      // No cost-based analysis except for Virtex-4, so no need to
      // call LabelNodes(), but we still want to hear about unrouted
      // nets.
      ReportUnroutedNets();
   }
}

// Report and count unrouted nets.
void NetChecker::ReportUnroutedNets() {
   Data::GroupIt groupIt(data.groupData.find("combined"));
   if (groupIt == data.groupData.end()) {
      FatalInternal(34, "Data from combined group not found.");
   }
   XCad_RMed* rmed(groupIt->second->GetRMed());
   CadFile::NetVec& netVec(groupIt->second->GetNets());
   for (CadFile::NetVecIt netVecIt(netVec.begin());
        netVecIt != netVec.end();
        ++netVecIt)
   {
      XCad_CadNetwork* net(*netVecIt);
      XCad_RMedRug* rug(rmed->GetRug(net));
      if (rug) {
         if (!rmed->IsFullyPlacedAndRouted(rug)) {
            Report("Input Error:44: Net "
                   + string(net->GetReportableName())
                   + " is not fully placed and routed.");
            if (!data.unroutedCount) {
               // Mention this the first time only.
               Report("  Isolation analysis cannot be completed without"
                      " complete routing\n"
                      "  information. Refer to the Place & Route Report"
                      " of your synthesis tool.");
            }
            ++data.unroutedCount;
         }
      }
   }
}

// R#report_interregion_signals - Label the nodes from the user's
// groups with the nets they are part of. Only the combined file
// is assumed to have routing information and be complete. No need to
// label nodes from the ignored arcs because we treat them as unused
// anyway.
void NetChecker::LabelNodes() {
   Data::GroupIt groupIt(data.groupData.find("combined"));
   if (groupIt == data.groupData.end()) {
      FatalInternal(34, "Data from combined group not found.");
   }
   XCad_RMed* rmed(groupIt->second->GetRMed());
   CadFile::NetVec& netVec(groupIt->second->GetNets());
   for (CadFile::NetVecIt netVecIt(netVec.begin());
        netVecIt != netVec.end();
        ++netVecIt)
   {
      XCad_CadNetwork* net(*netVecIt);
      XCad_RMedRug* rug(rmed->GetRug(net));
      if (rug) {
         for (const XCad_RMedRugNode* node(rug->GetFirstRugNode());
              node != NULL;
              node = node->GetNextRugNode())
         {
            Rf_Node rfNode(node->GetRfNode());
            // A node cannot appear in more than one net, so there is
            // no reasonable danger of overwriting something here.
            //
            // Net cannot be NULL because GetRug(net) would have
            // failed if net were NULL.
            nodeNet.insert(make_pair(rfNode, net));
         }
      } else if (net->IsConstant()) {
         // Is one of: Vcc, ground, driven by an unused block,
         // or driven by a block with a constant equation.
         ;
      } else {
         // Nets that are entirely within a comp will have no
         // RUGs because the routing with a comp is controlled by
         // the config string of the comp, not the general
         // interconnect, aka routing matrix.
         //
         // It would be possible to verify that this net was internal
         // by iterating over the parent blocks of its signals. If the
         // same parent is returned in all cases, we can certainly
         // ignore it. If the parents ever differ, then the rug
         // wouldn't have been null.
         //
         // I believe after studying it that this test is unnecessary
         // because the routing that leads to the comp containing
         // this net will be checked.
         ;
      }
   }
}

// R#report_errors - A node from arc must occur in net.
void NetChecker::CheckArcToNet(RevArc& arc, XCad_CadNetwork* net)
{
   XCad_RMed* rmed(data.groupData.find("combined")->second->GetRMed());
   XCad_RMedRug* rug(rmed->GetRug(net));
   ASSERT(rug);
   unsigned long sourceCount(0);
   unsigned long targetCount(0);
   for (const XCad_RMedRugNode* node(rug->GetFirstRugNode());
        node != NULL;
        node = node->GetNextRugNode())
   {
      Rf_Node rfNode(node->GetRfNode());
      if (rfNode == arc.GetSource()) {
         ++sourceCount;
      }
      if (rfNode == arc.GetTarget()) {
         ++targetCount;
      }
   }
   if (sourceCount == 0 && targetCount == 0) {
      Report("\nInternal Error:59: Net " + string(net->GetReportableName())
             + " has no nodes in common with ");
      ReportArc(arc);
      Report("Net " + string(net->GetReportableName())
             + " contains the following arcs: ");
      for (const XCad_RMedRugNode* node(rug->GetFirstRugNode());
           node != NULL;
           node = node->GetNextRugNode())
      {
         Rf_Arc rfArc(node->GetRfArc());
         Rf_Node rfNode(node->GetRfNode());
         if (rfArc && rfNode) {
            ReportArc(RevArc(rfArc, arc.GetOtherNode(rfNode)));
         }
         if (!rfNode) {
            Report("Internal Error:60: GetNext() returned undefined node.");
         }
      }
      Report("Contact Xilinx for assistance.");
   }
}

// R#report_errors - The target of the first node must be the source of
// the second, the target of the second node must be the source of the
// third, and so on.
void NetChecker::CheckArcToArc(Data::ArcPath& path) {
   Data::ArcPathIt prev(path.begin());
   for (Data::ArcPathIt apIt(prev); apIt != path.end(); ++apIt) {
      if (prev != apIt) {
         ASSERT(prev->GetTarget() == apIt->GetSource());
      }
      prev = apIt;
   }
}


// R#report_interregion_signals, R#report_errors - After this
// function, there will be a record of a fault from startNet to
// endNet. If the supplied path is shorter than a previous path for
// these nets, it will replace the previous path.
void NetChecker::AddFault(XCad_CadNetwork* startNet,
                          XCad_CadNetwork* endNet,
                          Data::ArcPath& path)
{
   // First some sanity checks.
   CheckArcToNet(path.front(), startNet);
   CheckArcToArc(path);
   CheckArcToNet(path.back(), endNet);

   // And now back to our show.
   ++faultCount;
   pair<XCad_CadNetwork*, XCad_CadNetwork*>
      key(make_pair(startNet, endNet));

   if (ArgParser::allPaths) {
      faultList.insert(make_pair(key, path));
   } else {
      PathIt prevIt(faultList.find(key));
      if (prevIt != faultList.end()) {
         // found another path from startNet to endNet
         if (path.size() < prevIt->second.size()) {
            // shorter than previous
            prevIt->second = path;
         }
      } else {
         // found a new path
         faultList.insert(make_pair(key, path));
      }
   }
}

// Follows the Unix tradition of returning 0 for success and something
// else for failure, in this case returning 1 for failure.
int NetChecker::GetStatus() {
   return faultCount == 0 ? 0 : 1;
}

// R#report_interregion_signals - Perform a depth first search where
// the depth is limited by the cost associated with the path.
//
// The cost is a lower bound on the number of faults that would be
// required for a signal path to be created.
//
// Recall an arc is a directed pair of nodes. If the target node is
// contained in a net in an isolation group distinct from the
// group parameter, and the total cost of the resulting path is less
// than the maximum allowed, a fault will be reported. Successors to
// arc will be explored recursively, until one of the following
// occurs:
//    1. the cost limit has been reached,
//    2. an isolation fault is found, or
//    3. the path connects to power or ground.
//
// Arcs that are already in the starting net or the ending net
// have zero cost because they are already configured.
//
// Invariants
//    - startNet is in group.
//    - On entry, the cost includes all arcs in path.
//    - The cost does not initially include the arc parameter.
//    - If path is empty, either the source or target of arc is in
//      group.
//    - The cost of the first arc in path is zero if the target is in
//      startNet.
void NetChecker::DepthLimitedSearch(XCad_CadNetwork* startNet,
                                    RevArc arc,
                                    const string& group,
                                    const int cost,
                                    Data::ArcPath& path)
{
   Rf_Node source(arc.GetSource());
   Rf_Node target(arc.GetTarget());

   string targetGroup("ignored");

   XCad_CadNetwork* sourceNet(0);
   XCad_CadNetwork* targetNet(0);

   NodeIt sourceIt(nodeNet.find(source));
   NodeIt targetIt(nodeNet.find(target));
   if (sourceIt != nodeNet.end()) {
      sourceNet = sourceIt->second;
   }
   if (targetIt != nodeNet.end()) {
      targetNet = targetIt->second;
      if (targetNet->IsVcc() || targetNet->IsGround()) return;
      targetGroup = data.netGroup.find(targetNet)->second;
   }

   if ((sourceNet == startNet || targetNet == startNet) && !path.empty()) {
      // The path visits the starting group more than once. Without
      // loss of generality, we can abort this search because every
      // arc in group will eventually be used as a starting point.
      return;
   }
   int newCost(cost);
   if (targetNet
       && targetGroup != "ignored"
       && targetGroup != "combined"
       && targetGroup != group)
   {
      // Complete path.
      if (sourceNet != targetNet) {
         newCost += costFunction->BitCount(arc);
         if (data.faultsMax <= newCost) return;
      } else {
         // No charge for arc because it is entirely
         // in the ending net.
         ;
      }
      path.push_back(arc);
      ++pathCount;
      AddFault(startNet, targetNet, path);
      path.pop_back();
   } else {
      if (targetNet == startNet) {
         // This is the first arc and the target is in the start net,
         // so there is no cost.
         ASSERT(path.empty());
      } else {
         newCost += costFunction->BitCount(arc);
         if (data.faultsMax <= newCost) return;
      }

      NodeCostMap::iterator ncIt(prevCost.find(target));
      if (ncIt == prevCost.end() || newCost < ncIt->second) {
         // Either we've never been here before by any route or none
         // of the previous routes were better than this one, so take
         // note!
         prevCost[target] = newCost;
      } else {
         // We have been here before, and the previous route was at
         // least as good. No need to continue from this node.
         return;
      }

      // Add it to the path and continue searching.
      path.push_back(arc);
      ++pathCount;

      Rf_ArcNodeContainer arcNodes(target.GetArcNodes());
      for (Rf_ArcNodeContainer::iterator arcNodesIt(arcNodes.begin());
           arcNodesIt != arcNodes.end();
           ++arcNodesIt)
      {
         RevArc nextArc(arcNodesIt->GetArc(), arc);
         DepthLimitedSearch(startNet, nextArc, group, newCost, path);
      }

      // At this point we've searched every extension of the path ending
      // with the arc.
      path.pop_back();
   }
}

// R#check_interconnect_routing - Explore potential routes through
// unrouted portions of a Virtex family device that could leak
// information from one isolated region to another.
//
// Search from every arc in every net except the "ignored" ones.
void NetChecker::Check() {
   Console(APPNAMESTR ":: Searching nets\n");

   // We'll need the rmed to find the arcs containing the node in a
   // net closest to the driver, also known as the root.
   XCad_RMed* rmed(data.groupData.find("combined")->second->GetRMed());
   Data::ArcPath path;

   // for each group... Note: the groups are sorted alphabetically and
   // there may be duplicates if a group is split across files.
   string prevGroup;
   for (Data::GroupIt groupIt(data.groupData.begin());
        groupIt != data.groupData.end();
        ++groupIt)
   {
      const string& group(groupIt->first);
      if (group != prevGroup) {
         // This map enables us to prune the search tree. It records
         // lowest cost to get to a node for a given group.
         prevCost.clear();      // Clear the cache for the new group.
      }
      prevGroup = group;
      for (Data::NetGroupIt netGroupIt(data.netGroup.begin());
           netGroupIt != data.netGroup.end();
           ++netGroupIt)
      {
         XCad_CadNetwork* net(netGroupIt->first);
         if (net->IsVcc() || net->IsGround()) {
            // Vcc and Gnd nets are exempt from isolation testing.
            continue;
         }
         // Search group-by-group so that prevCost is most effective.
         if (group != netGroupIt->second) {
            continue;
         }
         if (group == "ignored" || group == "combined") continue;
         ConsoleDot();

         XCad_RMedRug* rug(rmed->GetRug(net));
         if (rug) {
            for (const XCad_RMedRugNode* node(rug->GetFirstRugNode());
                 node != NULL;
                 node = node->GetNextRugNode())
            {
               Rf_Node rfNode(node->GetRfNode());
               Rf_ArcNodeContainer arcNodes(rfNode.GetArcNodes());
               for (Rf_ArcNodeContainer::iterator arcNodesIt(arcNodes.begin());
                    arcNodesIt != arcNodes.end();
                    ++arcNodesIt)
               {
                  RevArc nextArc(arcNodesIt->GetArc(), rfNode);
                  DepthLimitedSearch(net, nextArc, group, 0, path);
               }
            }
         } else {
            // This would be something we were warned about already.
         }
      }
   }
   Console("\n");
}

int NetChecker::GetErrorCount() {
   return faultCount;
}

// R#report_errors
void NetChecker::ReportArc(RevArc arc) {
   Rf_Node node0(arc.GetSource());
   Port_String tile0, wire0;
   Port::UINT32 wireIndex0(arc.GetSourceIndex());
   node0.GetWireNameAndTileInfo(wireIndex0, tile0, wire0);
   string tileName0(tile0);
   string wireName0(wire0);

   Rf_Node node1(arc.GetTarget());
   Port_String tile1, wire1;
   Port::UINT32 wireIndex1(arc.GetTargetIndex());
   node1.GetWireNameAndTileInfo(wireIndex1, tile1, wire1);
   string tileName1(tile1);
   string wireName1(wire1);

   string typeCode0(ToString(node0.GetTypeCode()));
   string typeCode1(ToString(node1.GetTypeCode()));

   NodeIt node0It(nodeNet.find(node0));
   NodeIt node1It(nodeNet.find(node1));

   string group0;
   if (node0It != nodeNet.end()) {
      group0 = data.netGroup.find(node0It->second)->second;
   } else {
      group0 = "unused";
   }

   string group1;
   if (node1It != nodeNet.end()) {
      group1 = data.netGroup.find(node1It->second)->second;
   } else {
      group1 = "unused";
   }

   string usage;

   if (group0 == group1) {
      usage = " is in group " + group0;
   } else {
      usage = " connects group " + group0 + " to group " + group1;
   }
   Report("  " + tileName0 + ":" + wireName0
          + ":" + typeCode0
          + " to " + tileName1 + ":" + wireName1
          + ":" + typeCode1
          + usage
          + " cost: " + ToString(costFunction->BitCount(arc))
          + " " + (arc.GetDirection() == RevArc::Forward
                   ? " forward" : " reverse"));
}


// R#report_interregion_signals, R#report_errors - Report a fault.
// Lists the starting net, the ending net, and the path between them
// expressed as command for FPGA Editor. The start net is colored
// yellow, the end net is colored blue, and the path is left selected
// (red). The path is composed of arcs and wires (known internally as
// nodes).
//
// A typical fault would be reported as:
//     FPGA Editor Script:
//       unselect -all
//       select net 'black2'
//       hilite -c yellow
//       unselect -all
//       select net 'red3'
//       hilite -c blue
//       zoom selection
//       unselect -all
//       select -k arc 'DUMMY(-712,1525)--->INPUT(-560,1160)'
//       select -k wire 'INPUT(-560,1160)'
//       select -k arc 'INPUT(-560,1160)--->OUTPUT(-560,1312)'
//       select -k wire 'OUTPUT(-560,1312)'
//       select -k arc 'OUTPUT(-560,1312)--->OMUX(-776,3457)'
//       select -k wire 'OMUX(-776,3457)'
//       select -k arc 'OUTPUT(-560,1608)--->OMUX(-776,3457)'
//
//     Nodes (Tile and Wire Names):
//       R4C4:YQ0
//       R4C5:OMUX_E2
//       R5C4:S2MID0
//       R4C5:E2MID1
//       R4C4:BX2
// This information can either be entered line-by-line into the
// command line of the FPGA Editor (using cut and paste), or written
// to a file with the extension .scr to be performed automatically.
void NetChecker::ReportFault(XCad_CadNetwork* startNet,
                             XCad_CadNetwork* endNet,
                             Data::ArcPath& path)
{
   Report("");
   if (startNet && endNet) {
      // Emit a heading if this is the first time or if either the start
      // or end group is different than last time.

      string& groupStart(data.netGroup.find(startNet)->second);
      string& groupEnd(data.netGroup.find(endNet)->second);
      if (!group1 || (*group1 != groupStart && *group2 != groupEnd)) {
         Report("Group " + groupStart + " to Group " + groupEnd + " Faults");
         Report("");
      }
      group1 = &groupStart;
      group2 = &groupEnd;
   } else {
      FatalInternal(80, "NetChecker::ReportFault called on null start or end"
                    " net.");
   }
   Report("FPGA Editor Script:");
   Report("  unselect -all");
   if (startNet) {
      Report("  select net '"
             + string(startNet->GetReportableName()) + "'");
      Report("  hilite -c yellow");
   }
   if (endNet) {
      Report("  unselect -all");
      Report("  select net '"
             + string(endNet->GetReportableName()) + "'");
      Report("  hilite -c blue");
      Report("  zoom selection");
   }
   Report("  unselect -all");
   {
      // We want to issue 'select wire' commands between the 'select
      // arc' commands, so the last iteration is different. We'll use
      // two iterators: one that starts from the first arc and one
      // that starts from the second. We use the second iterator to
      // end early.
      Data::ArcPathIt a(path.begin());
      if (a != path.end()) {
         Report("  select -k arc '" + string(a->GetRfArc().GetName()) + "'");
      } else {
         FatalInternal(28, "ReportFault called on empty path.");
      }
      Data::ArcPathIt b(path.begin());
      ++b;
      while (b != path.end()) {
         Report("  select -k wire '"
                + string(a->GetTarget().GetName())
                + "'");
         Report("  select -k arc '" + string(b->GetRfArc().GetName()) + "'");
         ++a;
         ++b;
      }
   }
   Report("");
   Report("Nodes by Tile and Wire Name:");
   for (Data::ArcPathIt a(path.begin()); a != path.end(); ++a) {
      ReportArc(*a);
   }
}

// Report a representative sample of isolation faults. One route is
// reported for each distinct ordered pair of nets that are
// insufficiently isolated.
void NetChecker::ReportFaults(bool verbose) {
   if (verbose) {

      Section("Cost-Based Net Isolation Violations");
      
      if (faultCount) {
         Report(
            "Isolation Violation:133: According to current fault costs,\n"
            "the paths below could be enabled with fewer than the required\n"
            "number of faults.");

         if (!ArgParser::allPaths) {
            Report(
               "  The list below is intended to be representative, not\n"
               "  complete. For each pair of isolated nets that cannot\n"
               "  be shown to be isolated, the path between them with\n"
               "  minimum cost is reported. Due to the relative abundance\n"
               "  of routing resources in the Virtex-4 FPGA, when any\n"
               "  failing paths exist, there are usually many variations\n"
               "  with equal cost.");
         }

         for (PathIt faultIt(faultList.begin());
              faultIt != faultList.end();
              ++faultIt)
         {
            ReportFault(faultIt->first.first,
                        faultIt->first.second,
                        faultIt->second);
         }
      } else {
         Report("  No paths between isolated nets can be enabled with\n"
                "  fewer than the minimum required faults.");
      }
   }

   Summary("\n  Net Isolation\n\n"
           "    Net Isolation Violations Found: " + ToString(faultCount));
   if (verbose) {
      Summary("    Net Isolation Violations Reported: "
              + ToString(faultList.size()));
   }
   Summary("    Routes Examined: " + ToString(pathCount));
}

// Path cost mode

// Search through all nodes of all the nets in the device for the
// one with the supplied name.
Rf_Node NetChecker::GetRoutedNodeByName(const char* name) {
   boost::shared_ptr<CadFile> cadFile(data.groupData.find("combined")->second);
   CadFile::NetVec& nets(cadFile->GetNets());
   XCad_RMed* rmed(cadFile->GetRMed());
   Port_String psName(name);
   for (CadFile::NetVecIt nIt(nets.begin());
        nIt != nets.end();
        ++nIt)
   {
      XCad_CadNetwork* net(*nIt);
      XCad_RMedRug* rug(rmed->GetRug(net));
      if (rug) {
         for (const XCad_RMedRugNode* node(rug->GetFirstRugNode());
              node != NULL;
              node = node->GetNextRugNode())
         {
            Rf_Node rfNode(node->GetRfNode());
            if (rfNode && rfNode.GetName() == psName) {
               return rfNode;
            }
         }
      }
   }
   const string wireName(name);
   const string deviceName(data.device);
   FatalInternal(6, "Unable to locate wire " + wireName + " in device "
                 + deviceName + ".");
   return Rf_Node();            // Not reached.
}

Rf_Arc NetChecker::GetArcFromNodeAndWireName(Rf_Node node,
                                             Port_String wireName)
{
   Rf_ArcNodeContainer arcNodes(node.GetArcNodes());
   for (Rf_ArcNodeContainer::iterator arcNodesIt(arcNodes.begin());
        arcNodesIt != arcNodes.end();
        ++arcNodesIt)
   {
      Rf_Arc arc(arcNodesIt->GetArc());
      pair<Rf_Node, Rf_Node> nodes(arc.GetNodes());
      if (nodes.first != node && nodes.second != node) {
         const string nodeName(node.GetName());
         const string arcName(arc.GetName());
         FatalInternal(38, "Inconsistent results from GetArcNodes().\n"
                       "Node "
                       + nodeName + " not connected to arc "
                       + arcName + ".");
      }
      if (arc.GetOtherNode(node).GetName() == wireName) {
         return arc;
      }
   }
   return Rf_Arc();             // Not found, return a bad one.
}

void NetChecker::AnalyzePathForward() {
   ASSERT(!data.scriptWires.empty());

   // Collect these so we can call ReportFault() at the end.
   Data::ArcPath path;
   XCad_CadNetwork* startNet(NULL);
   XCad_CadNetwork* endNet(NULL);

   // One of the arcs containing the first node is assumed to share a
   // node with a routed net of in the combined design.
   Rf_Node
      startNode(GetRoutedNodeByName(data.scriptWires.front().c_str()));
   Report("\nArcs connected to start node "
          + string(startNode.GetName())
          + ":");
   NodeIt startIt(nodeNet.find(startNode));
   if (startIt == nodeNet.end()) {
      const string nodeName(startNode.GetName());
      FatalInternal(79, "First node in path " + nodeName
                    + " is not part of any routed net.");
   }
   startNet = startIt->second;
   // For each of the remaining wire names, search the results of
   // GetArcNode() for the arc connecting prevNode to a node that
   // matches the wire name. Also keep track of the very last node, so
   // that the last net can be found also.
   Rf_Node lastNode;
   vector<string>::iterator wIt(data.scriptWires.begin());
   wIt++;
   for (Rf_Node node(startNode); wIt != data.scriptWires.end(); ++wIt) {
      Port_String psWire(wIt->c_str());
      Rf_Arc arc(GetArcFromNodeAndWireName(node, psWire));
      if (!arc) {
         const string nodeName(node.GetName());
         const string wireName(psWire);
         FatalInternal(12, "Node " + nodeName + " is not connected to node "
                       + wireName + ".");
      }
      path.push_back(RevArc(arc, node));
      node = arc.GetOtherNode(node);
      lastNode = node;

      // Find the ending net. lastNode and exactly one of the arcs
      // containing lastNode should be in a routed net.
      // No node before the last one should be included in the ending
      // net.
      Report("\nArcs connected to last node "
             + string(lastNode.GetName())
             + ":");
      NodeIt endIt(nodeNet.find(lastNode));
      if (endIt != nodeNet.end()) {
         endNet = endIt->second;
         Report(" * Found arc in net "
                + string(endNet->GetReportableName())
                + ".");
      } else {
         Report(" * None of the above arcs is routed.");
      }
   }
   ReportFault(startNet, endNet, path);
}

void NetChecker::AnalyzePathReverse() {
   reverse(data.scriptWires.begin(), data.scriptWires.end());
   AnalyzePathForward();
   reverse(data.scriptWires.begin(), data.scriptWires.end());
}

// Analyze the path the user supplied and report the cost for each
// wire therein.
void NetChecker::AnalyzePath() {
   AnalyzePathForward();
   AnalyzePathReverse();
}
