// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: main.c,v 1.62.206.3 2013/05/10 00:04:06 jcorbett Exp $
#include "main.h"

#include <iostream>
#include <locale>

#include "Portability/export/Port_MsgDispatcher.h"
#include "XOS/Portability/export/Port_Env.h"

#include "ConstraintsChecker.h"
#include "PinChecker.h"
#include "NetChecker.h"
#include "TileChecker.h"
#include "Data.h"
#include "ArgParser.h"

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.62.206.3 $";

// Helper function for wrapping a single line of text. See WordWrap()
// below. 
const string WordWrap1(const string text, size_t maxLineLen, size_t hang) {
   size_t nonSp(text.find_first_not_of(" "));
   size_t sp(text.find(" ", nonSp));
   if (nonSp == string::npos || text.size() <= maxLineLen || sp == string::npos) {
      return text;
   } else {
      size_t spPrev(sp);
      sp = text.find(" ", sp + 1);
      while (sp != string::npos && sp < maxLineLen) {
         spPrev = sp;
         sp = text.find(" ", sp + 1);
      }
      string head(text.substr(0, spPrev));
      if (spPrev == string::npos) {
         return head;
      } else {
         size_t tailNonSp(text.find_first_not_of(" ", spPrev + 1));
         if (tailNonSp == string::npos) {
            return head;
         } else {
            string tail(text.substr(tailNonSp));
            string margin(nonSp == string::npos ? 0 : nonSp + hang, ' ');
            return head + "\n" + WordWrap1(margin + tail, maxLineLen, 0);
         }
      }
   }
}

// Wrap long lines using a hanging indent of two characters.
const string WordWrap(const string text, size_t maxLineLen) {
   if (text.size()) {
      string result(text);
      size_t pos(result.find("\n"));
      string head(result.substr(0, pos));
      if (pos == string::npos) {
         return WordWrap1(head, maxLineLen, 2);
      } else {
         string tail(result.substr(pos + 1));
         return WordWrap1(head, maxLineLen, 2) + "\n" + WordWrap(tail, maxLineLen);
      }
   } else {
      return "";
   }
}

// Replace all instances of target in text with replacement and
// return the result.
const string StringReplaceAll(const string& text,
                                   const string& target,
                                   const string& replacement)
{
   string result(text);
   for (size_t p(result.find(target, 0));
        p != string::npos;
        p = result.find(replacement, p + replacement.size()))
   {
      result.replace(p, p + target.size(), replacement);
   }
   return result;
}

// Is prefix a prefix of candidate?
bool IsPrefix(const char* prefix, const char* candidate) {
   return strncmp(prefix, candidate, strlen(prefix)) == 0;
}

// Ugly hack to suppress warning about missing message translation
// file that we're not using anyhow.
void SuppressMissingMessageFileWarning() {
   Port_MsgDispatcher::getDefaultMsgMan()->disable();
   Port_LocStatus("");
   Port_MsgDispatcher::getDefaultMsgMan()->enable();
}

// The libraries that load device models and load design files send
// output to stdout. These messages are generally not helpful to end
// users. Constructing an instance of this class will suppress
// messages within the scope of the instance.

SuppressLibraryMessages::SuppressLibraryMessages() {
   Port_MsgDispatcher::getDefaultMsgMan()->disable();
}

SuppressLibraryMessages::~SuppressLibraryMessages() {
   Port_MsgDispatcher::getDefaultMsgMan()->enable();
}

// Originally, I didn't know how to suppress status output from Xilinx
// library commands, so for IVT output I used the Xilinx library
// output functions to avoid any difficulties resulting from using
// different output libraries simultaneously. I wrapped the Xilinx
// library output functions because they use variable length argument
// lists and thus allow errors that are not checked at compile time.
//
// In the transition from the I.32 build of the Xilinx libraries to
// the J.40 libraries, the Port_NOLOCStatus() function acquired a
// defect that corrupted output for long status messages. Meanwhile I
// had learned how to suppress Xilinx library output.
//
// This bug prompted me to change the wrapper to use the standard C++
// output operator, since it is now safe to do so.
//
// Flush after each output to make it easier to tune the timing of
// progress dots.
void Console(const string& s) {
   cout << s << flush;
}

void ConsoleDot(const string& s) {
   Console(s);
}

// Report an error. If the output is directed to the report file
// terminate execution immediately, otherwise buffer the message and
// allow the caller to continue.
void Fatal(const string& message) {
   Console(APPNAMESTR ":" + message + "\n");
   if (*ArgParser::errorOut) {
      *ArgParser::errorOut << message << endl;
   }
   if (ArgParser::errorOut == &ArgParser::reportOut) {
      // The message has been committed. Blow out of here.
      //
      // MS VC++ seems to incorrectly treat throw without an argument
      // as abort().
      throw false;
   } else {
      ;
      // Flow of control is to be determined by the caller. The error
      // has been added to the error buffer to be printed after the
      // header.
   }
}

// Report an error decorated with an error number, error type, and
// suffix. The prefix and suffix are optional and default to "Error: "
// and "\nContact Xilinx for assistance." respectively.
void Fatal(unsigned n,
           const string& message,
           const string& prefix,
           const string& suffix)
{
   const string fullText(prefix + ToString(n) + ": " + message + suffix);
   Fatal(fullText);
}

// Write a fatal internal error message and abort execution.
void FatalInternal(unsigned n, const string& s) {
   Fatal(n, s, "Internal Error:", "\nContact Xilinx for assistance.");
}

// Write a fatal input error message and abort execution.
void FatalInput(unsigned n, const string& s) {
   Fatal(n, s, "Input Error:", "");
}

// Write a fatal argument error message and abort execution.
void FatalUsage(unsigned n, const string& s) {
   Fatal(n, s, "Argument Error:", "");
}

// Write a string to the report file, if possible. Otherwise
// write to the console.
void Report(const string& s) {
   if (ArgParser::reportOut) {
      ArgParser::reportOut << s << endl;
   } else {
      Console(APPNAMESTR ":" + s + "\n");
   }
}

// Write a numbered warning.
void Warn(unsigned n, const string& s) {
   Report("Warning:" + ToString(n) + ": " + s);
}

// Write a numbered section heading and reset the subsection
// counter. Preincrement to simplify subsection numbering:
// GetSection() returns the number of the section most recently
// generated.
void Section(const string& s) {
   unsigned n(GetSectionNumber() + 1);
   SetSectionNumber(n);
   Report("\nSection " + ToString(n) + " - " + s + "\n");
   SetSubsectionNumber(0);
}

unsigned GetSectionNumber() {
   return ArgParser::sectionNumber;
}

void SetSectionNumber(unsigned n) {
   ArgParser::sectionNumber = n;
}

unsigned GetSubsectionNumber() {
   return ArgParser::subsectionNumber;
}

void SetSubsectionNumber(unsigned n) {
   ArgParser::subsectionNumber = n;
}

// Write a numbered subsection heading.
void Subsection(const string& s) {
   unsigned m(GetSubsectionNumber() + 1);
   SetSubsectionNumber(m);
   Report("\n  Section " + ToString(ArgParser::sectionNumber)
          + "." + ToString(m) + " - " + s + "\n");
}

// Append a string to the buffer used to hold the summary.
void Summary(const string& s) {
   ArgParser::summaryBuf << s + "\n";
}

// R#draw_fence
void Draw(const string& s) {
   ArgParser::drawingOut << s << "\n";
}

void EnableTestArcs();

// R#report_interregion_signals, R#check_irs_separation - There are
// more arcs in the hardware than are modeled for user designs. The
// environment variable below signals to that test arcs should be
// included when we list the arcs on a node.
//
// Moved from NetChecker constructor because TileChecker may need
// this, too.
void EnableTestArcs() {
   bool succeeded(!Port_Env::PutEnv("XIL_TEST_ARCS=1") &&
                  !Port_Env::PutEnv("XIL_DRM_EXCLUDE_ARCS=1") &&
                  !Port_Env::PutEnv("XIL_SPECIAL_FEATURES=1"));
   if (!succeeded) {
      FatalInternal(22, "Failed to enable test arcs.");
   }
}

int main(int argc, const char * const * argv) {
   SuppressMissingMessageFileWarning();
   try {
      Data data;
      ArgParser args(argc, argv, data);
      // Either we check the constraints file (.ucf) or we check the
      // placed and routed design (.ncd).
      int status(0);                     // 0 = success
      args.ReportHeader();               // R#report_provenance
      data.GetDeviceModel();             // Validate and cache
      if (args.NeedConstraintsCheck()) { // UCF mode
         ConstraintsChecker ucfChecker(data);
         TileChecker tileChecker(data);
         data.ReportSiteRanges();
         tileChecker.Check();
         tileChecker.GenerateDiagram();
         ucfChecker.Check();
         ucfChecker.ReportFaults(tileChecker);
         status = ucfChecker.GetStatus();
      } else {                           // NCD mode
         EnableTestArcs();
         NetChecker netChecker(data);
         PinChecker pinChecker(data);
         TileChecker tileChecker(data);
         data.CollectAreaRangesFromNCD();
         data.CollectBusMacros();
         data.LabelNets();
         data.ReportUserGlobalNets();
         if (args.verbose) {
            data.ReportUnusedNets();
            data.ReportBusMacroUsage();
            data.ReportSiteRanges();
         }
         tileChecker.Check();
         tileChecker.GenerateDiagram();
         tileChecker.ReportFaults();
         if (args.NeedPinCheck()) {
            pinChecker.Check();
            pinChecker.ReportFaults();
         } else {

            Summary("\n  Pin Isolation\n\n"
                    "    Pin analysis was not requested.");

         }
         ConsoleDot();
         Console("\n");
         if (args.NeedPathAnalysis()) {

            Section("Path Diagnostic Analysis");

            netChecker.AnalyzePath();
         } else {
            // Analysis to support net-based checking only exists
            // for Virtex-4. Subsequent devices are supported by
            // tile-based checking.
            if (data.GetDeviceModel()->IsVirtex4()) {
               netChecker.Check();
               netChecker.ReportFaults(args.verbose);
            }
         }

         string tiv(ToString(data.irsMultiLoadCount
                             + pinChecker.GetErrorCount()
                             + tileChecker.GetErrorCount()
                             + netChecker.GetErrorCount()));
         string un(ToString(data.unroutedCount));
         Summary("\n  NCD Isolation Verification Summary\n\n"
                 "    Total isolation violations: " + tiv + "\n"
                 "    Unrouted nets: " + un);

         status = !!
            (data.GetStatus()
             + pinChecker.GetStatus()
             + tileChecker.GetStatus()
             + netChecker.GetStatus());
      }

      if (status == 0) {
         Console("SUCCESS!\n");
      } else {
         Console("FAILURE.\n");
      }
      args.ReportFooter();
      ArgParser::CloseFiles();
      exit(status);
   } catch (bool) {
      ArgParser::CloseFiles();
      exit(-1);
   } catch (const std::exception& e) { // R#report_errors
      Report(e.what());
      ArgParser::CloseFiles();
      exit(-1);
   } catch (...) {                     // R#report_errors
      Report("Internal Error:70: Uncaught exception."
             "\nContact Xilinx for assistance.");
      ArgParser::CloseFiles();
      exit(-1);
   }
}
